/*
 * Main author(s) : Jean-Noël Monette
 * 
 * Copyright (c) 2013, Jean-Noël Monette
 * All rights reserved.

 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package convexPairs;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;


public class ConvexPairOfSum {

	private Parameters f;
	private int[] W;
	private int n;
	private PWLCF H;
	public boolean sort;
	
	public ConvexPairOfSum(Parameters f){
		this.f = f;
		this.n = f.getN();
		this.W = new int[n];
		this.H = null;
		this.sort = false;
	}

	private void computeHWithSort(){
		PWLCF H = new PWLCF();
		List<Segment> segsR = new ArrayList<Segment>();
		List<Segment> segsL = new ArrayList<Segment>();
		//The neutral of the function is given by some (minimal) assignment.
		H.x = 0;
		H.y = 0;
		//Visit all variables and add the corresponding segments.
		for(int i=0;i<n;i++){
			H.x += f.evaluateG(i, W[i]);
			H.y += f.evaluateF(i, W[i]);
			//right
			int k = W[i];
			int slope = f.getDeltaP(i, k);
			while(slope<Integer.MAX_VALUE){
				int nk = f.getBPP(i, k);
				int length = f.evaluateG(i,nk) - f.evaluateG(i,k);
				segsR.add(new Segment(slope,length));
				k = nk;
				slope = f.getDeltaP(i, k);
			}
			
			//left
			k  = W[i];
			slope = f.getDeltaN(i, k);
			while(slope<Integer.MAX_VALUE){
				int nk = f.getBPN(i, k);
				int length = f.evaluateG(i, k) - f.evaluateG(i, nk);
				segsL.add(new Segment(slope,length));
				k = nk;
				slope = f.getDeltaN(i, k);
			}
		}
		Collections.sort(segsL, new Comparator<Segment>() {
			public int compare(Segment o1, Segment o2) {
				return o1.slope - o2.slope;
			}
		});
		Collections.sort(segsR, new Comparator<Segment>() {
			public int compare(Segment o1, Segment o2) {
				return o1.slope - o2.slope;
			}
		});
		Segment cur = H.right;
		for(Segment s:segsR){
			cur = H.addSegment(s.slope,s.length,cur);
		}
		cur = H.left;
		for(Segment s:segsL){
			cur = H.addSegment(s.slope,s.length,cur);
		}
		H.close();
		this.H = H;
	}
	
	private void computeHByTraversal(){
		PWLCF H = new PWLCF();
		//The neutral of the function is given by some (minimal) assignment.
		H.x = 0;
		H.y = 0;
		//Visit all variables and add the corresponding segments.
		for(int i=0;i<n;i++){
			H.x += f.evaluateG(i, W[i]);
			H.y += f.evaluateF(i, W[i]);
			//right
			int k = W[i];
			int slope = f.getDeltaP(i, k);
			Segment cur = H.right;
			while(slope<Integer.MAX_VALUE){
				int nk = f.getBPP(i, k);
				int length = f.evaluateG(i,nk) - f.evaluateG(i,k);
				cur = H.addSegment(slope,length,cur);
				k = nk;
				slope = f.getDeltaP(i, k);
			}
			
			//left
			k  = W[i];
			slope = f.getDeltaN(i, k);
			cur = H.left;
			while(slope<Integer.MAX_VALUE){
				int nk = f.getBPN(i, k);
				int length = f.evaluateG(i, k) - f.evaluateG(i, nk);
				cur = H.addSegment(slope,length,cur);
				k = nk;
				slope = f.getDeltaN(i, k);
			}
		}
		H.close();
		this.H = H;
	}
	
	private class Seg_i {
		int u_0;
		int u_1;
		int l;
		int s;
		Segment segment;
	}
	
	private void getNextForwardSegment(int u_0, int i, Seg_i seg){
		seg.u_0 = u_0;
		seg.s = f.getDeltaP(i, seg.u_0);
		seg.u_1 = f.getBPP(i, seg.u_0);
		seg.l = f.evaluateG(i,seg.u_1) - f.evaluateG(i,seg.u_0);
	}
	private void getNextBackwardSegment(int u_0, int i, Seg_i seg){
		seg.u_0 = u_0;
		seg.s = f.getDeltaN(i, seg.u_0);
		seg.u_1 = f.getBPN(i, seg.u_0);
		seg.l = f.evaluateG(i,seg.u_0) - f.evaluateG(i,seg.u_1);
	}
	
	private void getNextLeftHSegment(Segment segment, Seg_i seg_rev_i, int i, Seg_i seg){
		int length_h = 0;
		int slope_h = 0;
		if(segment==null){
			length_h = Integer.MAX_VALUE;
			slope_h = Integer.MAX_VALUE;
		}else{
			length_h = segment.length;
			slope_h = segment.slope;
		}
		if(seg_rev_i.s < slope_h)throw new RuntimeException("Not possible");
		while(slope_h == seg_rev_i.s){
			if(slope_h==Integer.MAX_VALUE)break;
			length_h -= seg_rev_i.l; //remove from the current segment.
			if(length_h < 0)throw new RuntimeException("the length cannot be smaller than zero");
			getNextBackwardSegment(seg_rev_i.u_1, i,seg_rev_i);
			if(length_h==0){
				segment = segment.next;
				if(segment==null){
					length_h = Integer.MAX_VALUE;
					slope_h = Integer.MAX_VALUE;
				}else{
					length_h = segment.length;
					slope_h = segment.slope;
				}
			}
		}
		if(seg_rev_i.s < slope_h)throw new RuntimeException("Not possible");
		
		seg.u_0 = 0;//not used
		seg.s = slope_h;
		seg.u_1 = 0;//not used
		seg.l = length_h;
		seg.segment = segment;
	}
	
	private void getNextRightHSegment(Segment segment, Seg_i seg_rev_i, int i, Seg_i seg){
		int length_h = 0;
		int slope_h = 0;
		if(segment==null){
			length_h = Integer.MAX_VALUE;
			slope_h = Integer.MAX_VALUE;
		}else{
			length_h = segment.length;
			slope_h = segment.slope;
		}
		if(seg_rev_i.s < slope_h) throw new RuntimeException("Not possible");
		while(slope_h == seg_rev_i.s){
			if(slope_h==Integer.MAX_VALUE)break;
			length_h -= seg_rev_i.l; //remove from the current segment.
			if(length_h < 0)throw new RuntimeException("the length cannot be smaller than zero");
			getNextForwardSegment(seg_rev_i.u_1, i,seg_rev_i);
			if(length_h==0){
				segment = segment.next;
				if(segment==null){
					length_h = Integer.MAX_VALUE;
					slope_h = Integer.MAX_VALUE;
				}else{
					length_h = segment.length;
					slope_h = segment.slope;
				}
			}
		}
		if(seg_rev_i.s < slope_h)throw new RuntimeException("Not possible");
		
		seg.u_0 = 0;//not used
		seg.s = slope_h;
		seg.u_1 = 0;//not used
		seg.l = length_h;
		seg.segment = segment;
	}
	
	Seg_i seg_i = new Seg_i();
	Seg_i seg_rev_i = new Seg_i(); 
	Seg_i seg_h = new Seg_i();
	
	public void generalCaseForward(int i, int U_F, int L_G, int U_G){
		int curv = f.evaluateG(i,W[i]);
		int curhv = f.evaluateF(i, W[i]);
		int curb = H.x;
		int curHbv = H.y - f.evaluateF(i,W[i]);
		getNextForwardSegment(W[i], i, seg_i);
		getNextBackwardSegment(W[i], i, seg_rev_i);
		getNextLeftHSegment(H.left, seg_rev_i, i, seg_h);
		f.enforceIfGEqualThenFLessThanOrEqual(i, curv, U_F - curHbv);
		boolean move_b = curb < U_G && seg_h.s >= 0;
		while(seg_i.s < Integer.MAX_VALUE && (move_b || seg_h.s<Integer.MAX_VALUE) && curHbv +(move_b?0:seg_h.s)+ curhv + seg_i.s<= U_F){
			//move to the next
			curv += 1;
			curhv += seg_i.s;
			seg_i.l -= 1;
			if(move_b){
				curb+=1;
				//(b-v) stays constant.
			}else{
				curHbv += seg_h.s;
				seg_h.l-=1;
			}
			//constraint here
			f.enforceIfGEqualThenFLessThanOrEqual(i, curv, U_F - curHbv);
			if(seg_i.l==0){//reached end of segment for h(x_i)
				getNextForwardSegment(seg_i.u_1, i, seg_i);
			}
			if(seg_h.l==0){//reached end of segment for H
				getNextLeftHSegment(seg_h.segment.next, seg_rev_i, i, seg_h);
			}
			move_b = curb < U_G && seg_h.s >= 0;
		}
		f.enforceGLessThanOrEqual(i, curv);
	}
	
	public void specialCaseForward(int i, int U_F, int L_G, int U_G){
		int curv = f.evaluateG(i,W[i]);
		int curhv = f.evaluateF(i, W[i]);
		int curb = H.x;
		int curHbv = H.y - f.evaluateF(i,W[i]);
		getNextForwardSegment(W[i], i, seg_i);
		getNextBackwardSegment(W[i], i, seg_rev_i);
		getNextLeftHSegment(H.left, seg_rev_i, i, seg_h);
		boolean move_b = curb < U_G && seg_h.s >= 0;
		int length = Math.min(seg_h.l, Math.min(seg_i.l,move_b?U_G-curb:Integer.MAX_VALUE));
		while(seg_i.s < Integer.MAX_VALUE && (move_b || seg_h.s<Integer.MAX_VALUE) && curHbv +(move_b?0:length* seg_h.s)+ curhv + length*seg_i.s<= U_F){//end of the PWCLF, stop!
			curv += length;
			curhv += length * seg_i.s;
			seg_i.l -= length;
			if(move_b){
				curb +=length;
			}else{
				curHbv += length * seg_h.s;
				seg_h.l -= length;
			}
			if(seg_i.l==0){//reached end of segment for f(x_i)
				getNextForwardSegment(seg_i.u_1, i, seg_i);
			}
			if(seg_h.l==0){//reached end of segment for H
				getNextLeftHSegment(seg_h.segment.next, seg_rev_i, i, seg_h);
			}
			move_b = curb < U_G && seg_h.s >= 0;
			length = Math.min(seg_h.l, Math.min(seg_i.l,move_b?U_G-curb:Integer.MAX_VALUE));
		}
		//update curv, curHbv for the remaining length
		if(seg_i.s==Integer.MAX_VALUE || (!move_b && seg_h.s==Integer.MAX_VALUE)){
			length = 0;
		}else{
			length = Util.divFloor(U_F - curHbv - curhv, seg_i.s + (move_b?0:seg_h.s));
		}
		curv += length;
		if(!move_b && length > 0){
			if(seg_h.s==Integer.MAX_VALUE){
				System.out.println("Might have a problem");
				curHbv = Integer.MAX_VALUE;
			}else{
				curHbv += length*seg_h.s;
			}
		}
		f.enforceGLessThanOrEqual(i, curv);
		f.enforceIfGEqualThenFLessThanOrEqual(i, curv, U_F - curHbv);
	}
	
	public void generalCaseBackward(int i, int U_F, int L_G, int U_G){
		int curv = f.evaluateG(i,W[i]);
		int curhv = f.evaluateF(i, W[i]);
		int curb = H.x;
		int curHbv = H.y - f.evaluateF(i,W[i]);
		getNextBackwardSegment(W[i], i, seg_i);
		getNextForwardSegment(W[i], i, seg_rev_i);
		getNextRightHSegment(H.right, seg_rev_i, i, seg_h);
		f.enforceIfGEqualThenFLessThanOrEqual(i, curv, U_F - curHbv);
		boolean move_b = curb > L_G && seg_h.s >= 0;
		while(seg_i.s < Integer.MAX_VALUE && (move_b || seg_h.s<Integer.MAX_VALUE) && curHbv +(move_b?0:seg_h.s) + curhv + seg_i.s <= U_F){
			//move to the next
			curv -= 1;
			curhv += seg_i.s;
			seg_i.l -= 1;
			if(move_b){
				curb-=1;
				//(b-v) stays constant.
			}else{
				curHbv += seg_h.s;
				seg_h.l-=1;
			}	
			//constraint here
			f.enforceIfGEqualThenFLessThanOrEqual(i, curv, U_F - curHbv);
			if(seg_i.l==0){//reached end of segment for f(x_i)
				getNextBackwardSegment(seg_i.u_1, i, seg_i);
			}
			if(seg_h.l==0){//reached end of segment for H
				getNextRightHSegment(seg_h.segment.next, seg_rev_i, i, seg_h);
			}
			move_b = curb > L_G && seg_h.s >= 0;
		}
		f.enforceGGreaterThanOrEqual(i, curv);
	}
	
	public void specialCaseBackward(int i, int U_F, int L_G, int U_G){
		int curv = f.evaluateG(i,W[i]);
		int curhv = f.evaluateF(i, W[i]);
		int curb = H.x;
		int curHbv = H.y - f.evaluateF(i,W[i]);
		getNextBackwardSegment(W[i], i, seg_i);
		getNextForwardSegment(W[i], i, seg_rev_i);
		getNextRightHSegment(H.right, seg_rev_i, i, seg_h);
		boolean move_b = curb > L_G && seg_h.s >= 0;
		int length = Math.min(seg_h.l, Math.min(seg_i.l,move_b?curb-L_G:Integer.MAX_VALUE));
		while(seg_i.s < Integer.MAX_VALUE && (move_b || seg_h.s<Integer.MAX_VALUE) && curHbv +(move_b?0:length* seg_h.s)+ curhv + length*seg_i.s<= U_F){//end of the PWCLF, stop!
			curv -= length;
			curhv += length * seg_i.s;
			seg_i.l -= length;
			if(move_b){
				curb -=length;
			}else{
				curHbv += length * seg_h.s;
				seg_h.l -= length;
			}
			if(seg_i.l==0){//reached end of segment for f(x_i)
				getNextBackwardSegment(seg_i.u_1, i, seg_i);
			}
			if(seg_h.l==0){//reached end of segment for H
				getNextRightHSegment(seg_h.segment.next, seg_rev_i, i, seg_h);
			}
			move_b = curb > L_G && seg_h.s >= 0;
			length = Math.min(seg_h.l, Math.min(seg_i.l,move_b?curb-L_G:Integer.MAX_VALUE));
		}
		//update curv, curHbv for the remaining length
		if(seg_i.s==Integer.MAX_VALUE || (!move_b && seg_h.s==Integer.MAX_VALUE)){
			length = 0;
		}else{
			length = Util.divFloor(U_F - curHbv - curhv, seg_i.s + (move_b?0:seg_h.s));
		}
		curv -= length;
		if(!move_b && length > 0){
			if(seg_h.s==Integer.MAX_VALUE){
				System.out.println("Might have a problem");
				curHbv = Integer.MAX_VALUE;
			}else{
				curHbv += length*seg_h.s;
			}
		}
		f.enforceGGreaterThanOrEqual(i, curv);
		f.enforceIfGEqualThenFLessThanOrEqual(i, curv, U_F -curHbv);
	}

	public int computeU_G(int U_F){
		int curHb = H.y;
		int curb = H.x;
		Segment seg = H.right;
		int length = seg.length;
		int slope = seg.slope;
		while(slope < Integer.MAX_VALUE && curHb + length * slope <= U_F){
			curb +=length;
			curHb += length * slope;
			seg = seg.next;
			length = seg.length;
			slope = seg.slope;
		}
		if(slope == Integer.MAX_VALUE) return curb;
		else return curb + Util.divFloor(U_F - curHb,slope);
	}
	
	public int computeL_G(int U_F){
		int curHb = H.y;
		int curb = H.x;
		Segment seg = H.left;
		int length = seg.length;
		int slope = seg.slope;
		while(slope < Integer.MAX_VALUE && curHb + length * slope <= U_F){
			curb -=length;
			curHb += length * slope;
			seg = seg.next;
			length = seg.length;
			slope = seg.slope;
		}
		if(slope == Integer.MAX_VALUE) return curb;
		else return curb - Util.divFloor(U_F - curHb,slope);
	}
	
	public int computeLB(int L_G, int U_G) {
		for(int i = 0;i < n;i++){
			W[i] = f.getArgminF(i);
		}
		if(sort)computeHWithSort();
		else computeHByTraversal();
		if(H.x > U_G){
			int slack = H.x - U_G;
			Segment seg = H.left;
			int length = seg.length;
			int slope_max = seg.slope;
			while(slack - length > 0){
				slack -= length;
				seg = seg.next;
				length = seg.length;
				slope_max = seg.slope;
			}
			if(slope_max==Integer.MAX_VALUE && slack > 0){
				return Integer.MAX_VALUE;
			}
			for(int i = 0;i < n;i++){
				int slope = f.getDeltaN(i, W[i]);
				while(slope < slope_max){
					W[i] = f.getBPN(i,W[i]);
					slope = f.getDeltaN(i, W[i]);
				}
				if(slope == slope_max && slack > 0){
					int nw = Math.max(f.getBPN(i,W[i]),W[i] - slack);
					int l = f.evaluateG(i, W[i]) - f.evaluateG(i,nw);
					W[i] = nw;
					slack -= l;
				}
			}
			H.moveRootLeft(H.x-U_G);
		}else if(H.x < L_G){
			int slack = L_G - H.x;
			Segment seg = H.right;
			int length = seg.length;
			int slope_max = seg.slope;
			while(slack - length > 0){
				slack -= length;
				seg = seg.next;
				length = seg.length;
				slope_max = seg.slope;
			}
			if(slope_max==Integer.MAX_VALUE && slack > 0){
				return Integer.MAX_VALUE;
			}
			for(int i = 0;i < n;i++){
				int slope = f.getDeltaP(i, W[i]);
				while(slope < slope_max){
					W[i] = f.getBPP(i,W[i]);
					slope = f.getDeltaP(i, W[i]);
				}
				if(slope == slope_max && slack > 0){
					int nw = Math.min(f.getBPP(i,W[i]),W[i] + slack); 
					int l = f.evaluateG(i,nw) - f.evaluateG(i, W[i]);
					W[i] = nw;
					slack -= l;
				}
			}
			H.moveRootRight(L_G-H.x);
		}
		return H.y;
	}
}
